type ruban = Tete of char list * char * char list;;


let blancs = Tete ([], 'B', []);;
(*Les caractères au-delà des fins de listes étants blancs, il n'est pas nécessaire de remplir la liste pour que les caractères extérieurs soient considérés comme "blancs".*)

(** deplace_gauche
@param Tete (rub_g, car, rub_d), un ruban
@return un ruban identique au ruban d'entrée, mais avec la tête de lecture décalée d'un caractère vers la gauche.
*)
let deplace_gauche = fun (Tete (rub_g, car, rub_d)) -> match rub_g with
							| []	-> Tete ([], 'B', car::rub_d)
							| t::r 	-> Tete (r, t, car::rub_d);;
(* TEST :
# deplace_gauche (Tete(['1';'1'],'B',['1']));;
- : ruban = Tete (['1'], '1', ['B'; '1'])
*)

(** deplace_droite
@param Tete (rub_g, car, rub_d), un ruban
@return un ruban identique au ruban d'entrée, mais avec la tête de lecture décalée d'un caractère vers la droite.
*)
let deplace_droite = fun (Tete (rub_g, car, rub_d)) -> match rub_d with
							| []	-> Tete (car::rub_g, 'B', [])
							| t::r	-> Tete (car::rub_g, t, r);;
(* TEST :
# deplace_droite (Tete(['1'],'B',['1';'1']));;
- : ruban = Tete (['B'; '1'], '1', ['1'])
*)

(** charList_to_string
@param l, une liste de char
@return une chaine de caractère composée de la concaténation des char de la liste, dans l'ordre et séparés par des espaces.
*)
let charList_to_string = fun l -> List.fold_right (fun car str -> String.make 1 car ^ " " ^ str) l "";;
(* TEST :
# charList_to_string ['a';'b';'c';'d';'e'];;
- : string = "a b c d e "
# charList_to_string [];;
- : string = ""
*)

(** inv_liste
@param l, une liste
@return une liste contenant les même éléments que l, mais dans l'ordre inverse.
*)
let inv_liste = fun l -> List.fold_left (fun r elt -> elt::r) [] l;;
(* TEST :
# inv_liste [1;2;3;4;5];;
- : int list = [5; 4; 3; 2; 1]
*)

(** affiche_ruban
@param Tete (rub_g, car, rub_d), un ruban
@return un type unit, et affiche le ruban sur la sortie standard
*)
let affiche_ruban = fun (Tete (rub_g, car, rub_d)) ->
	Printf.printf "%s%c%c%c %s" (charList_to_string (inv_liste rub_g))  '[' car ']' (charList_to_string rub_d);;
(* TEST :
# affiche_ruban blancs;;
[B] - : unit = ()
# affiche_ruban (Tete(['1';'1';'B';'1'],'B',['B';'1']));;
1 B 1 1 [B] B 1 - : unit = ()
*)

(** rechercher
@param liste, une liste de couples (clef, valeur)
@param clef, une clef à rechercher la liste
@return une erreur "Not_found" si la clef n'est pas trouvée, la valeur associée à la clef sinon.
*)
let rec rechercher = fun liste clef -> match liste with
					| [] -> failwith "Not_found"
					| (key, value)::r -> if key = clef then value else rechercher r clef;;

(* TEST :
# rechercher [('a',1);('b',2);('c',3)] 'b';;
- : int = 2
*)

type direction = G | D | N;;
(*le type direction est défini par trois constante : G pour gauche, D pour droite et N pour l'absence d direction (utile pour la machine à deux rubans)*)

type 'a turing_machine = Machine of 'a * (('a*char)*('a*char*direction)) list * 'a;;

(** pas
@param Machine (etat_init, liste_assoc, etat_final), une machine de Turing
@param Tete (rub_g, car_tete, rub_d), un ruban
@param etat, l'état actuel de la machine
@return le couple composé du nouvel état et du ruban obtenu après avoir effectué une étape de transition de la machine de turing en partant de l'état et du ruban donné
*)
let pas = fun (Machine (etat_init, liste_assoc, etat_final)) (Tete (rub_g, car_tete, rub_d)) etat ->
	let trans = rechercher liste_assoc (etat, car_tete) in
	match trans with
	| (nouv_etat, nouv_car, G) -> (nouv_etat, deplace_gauche (Tete (rub_g, nouv_car, rub_d)))
	| (nouv_etat, nouv_car, D) -> (nouv_etat, deplace_droite (Tete (rub_g, nouv_car, rub_d)))
	| (nouv_etat, nouv_car, N) -> (nouv_etat, (Tete (rub_g, nouv_car, rub_d)));;
	(*Ce dernier cas n'est jamais utilisé, mais le pattern matching doit être exhaustif.*)
(* TEST :
Les tests de pas se font avec ceux de execute.
*)

(** execute
@param Machine (etat_init, liste_assoc, etat_final), une machine de Turing
@param rub, un ruban
@return le ruban une fois que la machine a atteint son état final.
*)
let execute = fun (Machine (etat_init, liste_assoc, etat_final)) rub ->
	affiche_ruban rub ; print_newline ();
	let rec execute_rec = fun machine rub etat ->
		if etat != etat_final then let (nouv_etat, nouv_rub) = pas machine rub etat in
			affiche_ruban nouv_rub ; print_newline () ;
			execute_rec machine nouv_rub nouv_etat
		else rub
	in execute_rec (Machine (etat_init, liste_assoc, etat_final)) rub etat_init;;
(* TEST :
# execute castor_affaire blancs;;
[B] 
1 [B] 
[1] 1 
[B] 1 1 
[B] 1 1 1 
[B] 1 1 1 1 
1 [1] 1 1 1 
1 1 [1] 1 1 
1 1 1 [1] 1 
1 1 1 1 [1] 
1 1 1 1 1 [B] 
1 1 1 1 [1] 1 
1 1 1 [1] 1 1 
1 1 1 1 [1] 1 
- : ruban = Tete (['1'; '1'; '1'; '1'], '1', ['1'])

# execute mystere (Tete([],'B',[]));;
[B] 
B [B] 
- : ruban = Tete (['B'], 'B', [])
# execute mystere (Tete([],'1',[]));;
[1] 
B [B] 
B B [B] 
B [B] 1 
[B] B 1 
1 [B] 1 
1 B [1] 
- : ruban = Tete (['B'; '1'], '1', [])

# execute mystere (Tete(['1'],'1',[]));;
1 [1] 
1 B [B] 
1 B B [B] 
1 B [B] 1 
1 [B] B 1 
1 1 [B] 1 
1 1 B [1] 
- : ruban = Tete (['B'; '1'; '1'], '1', [])

# execute mystere (Tete([],'1',['1']));;
[1] 1 
B [1] 
B 1 [B] 
B 1 B [B] 
B 1 [B] 1 
B [1] B 1 
[B] 1 B 1 
1 [1] B 1 
1 B [B] 1 
1 B B [1] 
1 B B 1 [B] 
1 B B [1] 1 
1 B [B] 1 1 
1 [B] B 1 1 
1 1 [B] 1 1 
1 1 B [1] 1 
- : ruban = Tete (['B'; '1'; '1'], '1', ['1'])

# execute mystere (Tete(['1'],'1',['1']));;
1 [1] 1 
1 B [1] 
1 B 1 [B] 
1 B 1 B [B] 
1 B 1 [B] 1 
1 B [1] B 1 
1 [B] 1 B 1 
1 1 [1] B 1 
1 1 B [B] 1 
1 1 B B [1] 
1 1 B B 1 [B] 
1 1 B B [1] 1 
1 1 B [B] 1 1 
1 1 [B] B 1 1 
1 1 1 [B] 1 1 
1 1 1 B [1] 1 
- : ruban = Tete (['B'; '1'; '1'; '1'], '1', ['1'])
*)

let castor_affaire = Machine ('a', [	(('a','B'),('b','1',D));
					(('a','1'),('c','1',G));
					(('b','B'),('a','1',G));
					(('b','1'),('b','1',D));
					(('c','B'),('b','1',G));
					(('c','1'),('h','1',D))
				], 'h');;

let mystere = Machine (1, [	((1,'B'),(0,'B',D));
				((1,'1'),(2,'B',D));
				((2,'B'),(3,'B',D));
				((2,'1'),(2,'1',D));
				((3,'B'),(4,'1',G));
				((3,'1'),(3,'1',D));
				((4,'B'),(5,'B',G));
				((4,'1'),(4,'1',G));
				((5,'B'),(1,'1',D));
				((5,'1'),(5,'1',G));
			], 0);;

(*
La machine mystère (sur un ruban contenant uniquement n '1') :
La machine mystère génère le ruban composé de n '1', un caractère blanc, et n-k '1' (avec k la position initiale de la tête), avec la tête positionnée juste à droite du caractère blanc.
*)


type 'a turing_machine2 = Machine2 of 'a * (('a*char*char)*('a*char*char*direction*direction)) list * 'a;;
(*Le type turing_machine2 est très semblable à turing_machine, mais elle associe un état, deux caractères et deux direction à un état et deux directions.
Par ailleurs, le type 'a sera un couple des types d'état des deux machines traitant chacune un ruban.*)

(** pas2
@param Machine2 (etat_init, liste_assoc, etat_final),une machine de turing à deux ruban
@param Tete (rub_g1, car_tete1, rub_d1), un premier ruban
@param Tete (rub_g2, car_tete2, rub_d2), un second ruban
@param etat, l’état actuel de la machine
@return un couple (nouvel état, (nouveau ruban 1, nouveau ruban 2))
*)
let pas2 = fun (Machine2 (etat_init, liste_assoc, etat_final)) (Tete (rub_g1, car_tete1, rub_d1)) (Tete (rub_g2, car_tete2, rub_d2)) etat ->
	let trans = rechercher liste_assoc (etat, car_tete1, car_tete2) in
	match trans with
	| (nouv_etat, nouv_car1, nouv_car2, G, G) -> (nouv_etat, deplace_gauche (Tete (rub_g1, nouv_car1, rub_d1)), deplace_gauche (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, G, D) -> (nouv_etat, deplace_gauche (Tete (rub_g1, nouv_car1, rub_d1)), deplace_droite (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, G, N) -> (nouv_etat, deplace_gauche (Tete (rub_g1, nouv_car1, rub_d1)), (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, D, G) -> (nouv_etat, deplace_droite (Tete (rub_g1, nouv_car1, rub_d1)), deplace_gauche (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, D, D) -> (nouv_etat, deplace_droite (Tete (rub_g1, nouv_car1, rub_d1)), deplace_droite (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, D, N) -> (nouv_etat, deplace_droite (Tete (rub_g1, nouv_car1, rub_d1)), (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, N, G) -> (nouv_etat, (Tete (rub_g1, nouv_car1, rub_d1)), deplace_gauche (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, N, D) -> (nouv_etat, (Tete (rub_g1, nouv_car1, rub_d1)), deplace_droite (Tete (rub_g2, nouv_car2, rub_d2)))
	| (nouv_etat, nouv_car1, nouv_car2, N, N) -> (nouv_etat, (Tete (rub_g1, nouv_car1, rub_d1)), (Tete (rub_g2, nouv_car2, rub_d2)));;
	(*Ce dernier cas ne devrait pas apparaître, car il signifie que les deux sous-machines à un ruban de la machine à deux rubans ont terminé, ce qui correspond à la terminaison de la machine à deux rubans, mais le pattern matching doit être exhaustif.*)
(* TEST :
Les tests de pas2 se font avec ceux de execute2.
*)

(** execute2
@param une machine à deux rubans
@param rub1, le premier ruban
@param rub2, le second ruban
@return le couple des deux rubans une fois que la machine a atteint son état final
*)
let execute2 = fun (Machine2 (etat_init, liste_assoc, etat_final)) rub1 rub2 ->
	affiche_ruban rub1 ; print_string "| " ; affiche_ruban rub2 ; print_newline () ;
	let rec execute_rec2 = fun machine rub1 rub2 etat ->
		let ((e1,e2),(ef1,ef2)) = (etat,etat_final) in
		if e1 != ef1 || e2 != ef2 then let (nouv_etat, nouv_rub1, nouv_rub2) = pas2 machine rub1 rub2 etat in
			affiche_ruban nouv_rub1 ; print_string "| " ; affiche_ruban rub2 ; print_newline () ;
			execute_rec2 machine nouv_rub1 nouv_rub2 nouv_etat
		else (rub1, rub2)
	in execute_rec2 (Machine2 (etat_init, liste_assoc, etat_final)) rub1 rub2 etat_init;;
(* TEST :
# execute2 machine_double blancs (Tete(['1'],'1',['1']));;
[B] | 1 [1] 1 
1 [B] | 1 [1] 1 
[1] 1 | 1 B [1] 
[B] 1 1 | 1 B 1 [B] 
[B] 1 1 1 | 1 B 1 B [B] 
[B] 1 1 1 1 | 1 B 1 [B] 1 
1 [1] 1 1 1 | 1 B [1] B 1 
1 1 [1] 1 1 | 1 [B] 1 B 1 
1 1 1 [1] 1 | 1 1 [1] B 1 
1 1 1 1 [1] | 1 1 B [B] 1 
1 1 1 1 1 [B] | 1 1 B B [1] 
1 1 1 1 [1] 1 | 1 1 B B 1 [B] 
1 1 1 [1] 1 1 | 1 1 B B [1] 1 
1 1 1 1 [1] 1 | 1 1 B [B] 1 1 
1 1 1 1 [1] 1 | 1 1 [B] B 1 1 
1 1 1 1 [1] 1 | 1 1 1 [B] 1 1 
- : ruban * ruban =
(Tete (['1'; '1'; '1'; '1'], '1', ['1']),
 Tete (['B'; '1'; '1'; '1'], '1', ['1']))
*)

(** parallelise
@param deux machines de Turing à un ruban
@return une machine de Turing à deux rubans composée des deux machines à un rubanfournies en paramètre
*)
let parallelise = fun (Machine (etat_init1, liste_assoc1, etat_final1)) (Machine (etat_init2, liste_assoc2, etat_final2)) ->
	(*distribue_assoc prend en paramètre une liste d'associations pour une machine à un ruban et une association, et renvoie une liste d'asociations pour une turing_machine2.
	La liste d'association renvoyée est en fait la liste d'entrée sur laquelle a été "distribuée" l'association donnée en paramètre.*)
	let rec distribue_assoc = fun liste_assoc ((etat_init, car),(nouv_etat, nouv_car, nouv_dir)) ->
		match liste_assoc with
		| [] -> []
		| ((etat_initL,carL),(nouv_etatL,nouv_carL,nouv_dirL))::reste -> (((etat_initL,etat_init),carL,car),((nouv_etatL,nouv_etat),nouv_carL,nouv_car,nouv_dirL,nouv_dir))::(distribue_assoc reste ((etat_init, car),(nouv_etat, nouv_car, nouv_dir)))
	in
	(*distribue_liste_assoc prend en paramètre deux listes d'associations de machine à un ruban, et renvoie la liste d'associations "fusionnée" des deux pour une machine à deux ruban.
	Pour chaque élément de la deuxième liste, distribue_liste_assoc appelle distribue_assoc sur la première liste et l'élément en cours.*)
	let rec distribue_liste_assoc = fun liste1 liste2 ->
		match liste2 with
		| [] -> []
		| assoc::reste -> (distribue_assoc liste1 assoc)@(distribue_liste_assoc liste1 reste)
	in
	(*on renvoie la machine dont la liste d'association est l'ensemble des combinaisons d'association des deux machines,
	et à laquelle on ajoute quatre association pour rester dans l'état final d'une machine lorsque l'une des deux termine avant l'autre.*)
	Machine2 ((etat_init1, etat_init2), distribue_liste_assoc
				(((etat_final1,'1'),(etat_final1,'1',N))::((etat_final1,'B'),(etat_final1,'B',N))::liste_assoc1)
				(((etat_final2,'1'),(etat_final2,'1',N))::((etat_final2,'B'),(etat_final2,'B',N))::liste_assoc2),
				(etat_final1, etat_final2));;
(* TEST :
# execute2 machine_double blancs (Tete(['1'],'1',['1']));;
[B] | 1 [1] 1 
1 [B] | 1 [1] 1 
[1] 1 | 1 B [1] 
[B] 1 1 | 1 B 1 [B] 
[B] 1 1 1 | 1 B 1 B [B] 
[B] 1 1 1 1 | 1 B 1 [B] 1 
1 [1] 1 1 1 | 1 B [1] B 1 
1 1 [1] 1 1 | 1 [B] 1 B 1 
1 1 1 [1] 1 | 1 1 [1] B 1 
1 1 1 1 [1] | 1 1 B [B] 1 
1 1 1 1 1 [B] | 1 1 B B [1] 
1 1 1 1 [1] 1 | 1 1 B B 1 [B] 
1 1 1 [1] 1 1 | 1 1 B B [1] 1 
1 1 1 1 [1] 1 | 1 1 B [B] 1 1 
1 1 1 1 [1] 1 | 1 1 [B] B 1 1 
1 1 1 1 [1] 1 | 1 1 1 [B] 1 1 
- : ruban * ruban =
(Tete (['1'; '1'; '1'; '1'], '1', ['1']),
 Tete (['B'; '1'; '1'; '1'], '1', ['1']))
*)
